/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: cmd_mode.c,v 1.16.4.11 2002/04/02 11:33:18 lloydy Exp $
 *
 */
#include "chunkyircd.h"
#include "network.h"
#include "ban.h"

CMIRCD_MODULE(1.02, CMD_CLIENT | CMD_SERVER);

/* Routines in this file, not available globally */
static int chan_mode_change(ServerInfo, ClientInfo, ChannelInfo, Cmd);
static int chan_mode_report(ClientInfo, ChannelInfo);
static int list_bans(ClientInfo, ChannelInfo);
static int check_mode_ban(char *);
static int client_mode(ServerInfo, ClientInfo, char *);
static int client_mode_return(ClientInfo);
static void chan_mode_change_with_param(ClientInfo const commander,
					ChannelInfo const chan,
					const int givetake,
					const char mode_char,
					const char *param, int *curr_param,
					char *mode, int *mode_pos,
					char *params);
static void chan_mode_change_without_param(ChannelInfo const chan,
					   const int givetake,
					   const char mode_char,
					   char *mode, int *mode_pos,
					   char *params, int *param_len);

/* 
 * cmd_mode - deals with MODE commands 
 */
int cmd_mode(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
	ChannelInfo chan = NULL;
	ChannelMember chan_user = NULL;
	ClientInfo client = NULL;
	char buffer[BUF_SIZE];

	/* Check we have at least 1 params */
	if (!check_params(cmd, 1)) {
		/* Return Error */
		send_numeric(c_ptr, get_my_server(), 461,
			     "MODE :Not enough params");

		return 1;
	}

	/* Server ones */
	if ((s_ptr) && (cmd->origin)
	    && (client = client_locate(cmd->origin))) {
		if ((chan = locate_channel(cmd->params[0]))) {
			chan_mode_change(s_ptr, client, chan, cmd);
		} else {
			client_mode(s_ptr, client, cmd->params[1]);
		}

		return 1;
	}
	if (s_ptr) {

		return 0;
	}

	/* Is the first param a channel? */
	chan = locate_channel(cmd->params[0]);

	if (chan != NULL) {
		/* its a channel */
		if (cmd->params[1] != NULL) {
			/* Mode changes */
			/* Check for ops */
			chan_user = channel_is_client_on(c_ptr, chan);
			if (chan_user == NULL) {
				/* Not on channel */
				ircd_sprintf(buffer, BUF_SIZE,
					     "%s : You're not on that channel",
					     chan->name);
				send_numeric(c_ptr, get_my_server(), 442,
					     buffer);

				return 1;
			}
			if (!(chan_user->chan_mode & CHANUSR_OP)
			    && !(c_ptr->mode & MODE_IRCOP)
			    && !(chan_user->chan_mode & CHANUSR_HALFOP)) {
				/* Not got ops */
				/* Check to see if it's a ban command */
				if (check_mode_ban(cmd->params[1])) {
					list_bans(c_ptr, chan);
					return 1;
				} else {
					ircd_sprintf(buffer, BUF_SIZE,
						     "%s :Ha! You're no chan op",
						     chan->name);
					send_numeric(c_ptr,
						     get_my_server(), 482,
						     buffer);

					return 1;
				}
			}
			chan_mode_change(s_ptr, c_ptr, chan, cmd);
		} else {
			/* Report Current mdoe */
			chan_mode_report(c_ptr, chan);
			ircd_sprintf(buffer, BUF_SIZE, "%s %ld",
				     chan->name, chan->ts);
			send_numeric(c_ptr, get_my_server(), 329, buffer);
		}
	} else {
		client = client_locate(cmd->params[0]);
		/* Must be a client */
		if (!client) {
			/* Invalid nick */
			ircd_sprintf(buffer, BUF_SIZE, "%s :No such nick",
				     cmd->params[0]);
			send_numeric(c_ptr, get_my_server(), 401, buffer);

			return 0;
		}
		if (client != c_ptr) {
			/* trying to chnage somebody else's modes */
			send_numeric(c_ptr, get_my_server(), 502,
				     ":Cannot change mode for other users");

			return 0;
		}
		/* We can do what we want */
		if (cmd->params[1] != NULL) {
			/* Change mode */
			client_mode(NULL, client, cmd->params[1]);
		} else {
			/* Return Current Mode */
			client_mode_return(client);
		}

	}



	return 1;
}

/*
 * Deals with a mode change for a mode with a parameter
 */
static void chan_mode_change_with_param(ClientInfo const commander,
					ChannelInfo const chan,
					const int givetake,
					const char mode_char,
					const char *in_param,
					int *curr_param, char *mode,
					int *mode_pos, char *params)
{
	ClientInfo client = NULL;
	ChannelMember chan_member = NULL;
	ChannelMember commander_check = NULL;
	int params_len;

	switch (mode_char) {

		/* Channel Operator */
	case 'o':
		client = client_locate(in_param);
		chan_member = channel_is_client_on(client, chan);
		(*curr_param)++;

		/* Check that the subject is on the channel */
		if (!chan_member || !givetake) {
			return;
		}

		commander_check = channel_is_client_on(commander, chan);
		/* Check that this isn't a halfop setting +o */
		if (!(commander->mode & MODE_IRCOP) &&
		    !(commander_check
		      && commander_check->chan_mode & CHANUSR_OP)) {
			char err_buffer[BUF_SIZE];
			ircd_sprintf(err_buffer, BUF_SIZE,
				     "%s :You're not channel operator",
				     chan->name);
			send_numeric(commander, get_my_server(), 482,
				     err_buffer);
			return;
		}

		/* Set the mode */
		if (givetake > 0) {
			if (!(chan_member->chan_mode & CHANUSR_OP)) {
				/* Set mode, and put in out mode */
				chan_member->chan_mode |= CHANUSR_OP;
				mode[*mode_pos] = 'o';
				(*mode_pos)++;

				/* Add nick to the params */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		} else {
			if (chan_member->chan_mode & CHANUSR_OP) {
				chan_member->chan_mode &= ~CHANUSR_OP;
				mode[*mode_pos] = 'o';
				(*mode_pos)++;

				/* Add Nick */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		}
		break;

		/* TS4 Halfop */
	case 'h':
		client = client_locate(in_param);
		chan_member = channel_is_client_on(client, chan);
		(*curr_param)++;

		/* Check that the subject is on the channel */
		if (!chan_member || !givetake) {
			return;
		}

		/* Set the mode */
		if (givetake > 0) {
			if (!(chan_member->chan_mode & CHANUSR_HALFOP)) {
				/* Set mode, and put in out mode */
				chan_member->chan_mode |= CHANUSR_HALFOP;
				mode[*mode_pos] = 'h';
				(*mode_pos)++;

				/* Add nick to the params */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		} else {
			if (chan_member->chan_mode & CHANUSR_HALFOP) {
				chan_member->chan_mode &= ~CHANUSR_HALFOP;
				mode[*mode_pos] = 'h';
				(*mode_pos)++;

				/* Add Nick */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		}
		break;

		/* Channel Voice */
	case 'v':
		client = client_locate(in_param);
		chan_member = channel_is_client_on(client, chan);
		(*curr_param)++;

		/* Check that the subject is on the channel */
		if (!chan_member || !givetake) {
			return;
		}

		/* Set the mode */
		if (givetake > 0) {
			if (!(chan_member->chan_mode & CHANUSR_VOICE)) {
				/* Set mode, and put in out mode */
				chan_member->chan_mode |= CHANUSR_VOICE;
				mode[*mode_pos] = 'v';
				(*mode_pos)++;

				/* Add nick to the params */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		} else {
			if (chan_member->chan_mode & CHANUSR_VOICE) {
				chan_member->chan_mode &= ~CHANUSR_VOICE;
				mode[*mode_pos] = 'v';
				(*mode_pos)++;

				/* Add Nick */
				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, client->nick,
					300 - params_len);
			}
		}
		break;

		/* Channel bans */
	case 'b':
		if (!in_param) {
			list_bans(commander, chan);
		} else {
			if (givetake > 0) {
				char *ban_retval = NULL;
				ban_retval = add_ban(chan, in_param);

				if (ban_retval) {
					params_len = strlen(params);
					strncat(params, " ",
						300 - params_len);
					params_len++;
					strncat(params, ban_retval,
						300 - params_len);

					mode[*mode_pos] = 'b';
					(*mode_pos)++;
					(*curr_param)++;
				}
			}
			if (givetake < 0) {
				if (rem_ban(chan, in_param)) {
					params_len = strlen(params);
					strncat(params, " ",
						300 - params_len);
					params_len++;
					strncat(params, in_param,
						300 - params_len);

					mode[*mode_pos] = 'b';
					(*mode_pos)++;
					(*curr_param)++;
				}
			}
		}
		break;

		/* Channel Limit */
	case 'l':

		if (givetake > 0) {
			/* Set the limit */
			int new_limit;
			char limit_str[20];

			new_limit = atoi(in_param);
			(*curr_param)++;

			chan->user_limit = new_limit;
			mode[*mode_pos] = 'l';
			(*mode_pos)++;

			params_len = strlen(params);
			strncat(params, " ", 300 - params_len);
			params_len++;
			ircd_sprintf(limit_str, 20, "%d", new_limit);
			strncat(params, limit_str, 300 - params_len);
		}

		if (givetake < 0 && chan->user_limit != -1) {
			chan->user_limit = -1;
			mode[*mode_pos] = 'l';
			(*mode_pos)++;
		}

		break;

		/* Channel Keys */
	case 'k':
		if (in_param) {
			if (givetake > 0) {
				if (chan->key) {
					free(chan->key);
					chan->key = NULL;
				}

				chan->key = string_alloc(strlen(in_param));

				strcpy(chan->key, in_param);

				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, chan->key,
					300 - params_len);

				mode[*mode_pos] = 'k';
				chan->mode |= CMODE_KEY;
				(*mode_pos)++;
				(*curr_param)++;
			}
			if (givetake < 0
			    && strcmp(in_param, chan->key) == 0) {

				free(chan->key);
				chan->key = NULL;
				chan->mode &= ~CMODE_KEY;

				params_len = strlen(params);
				strncat(params, " ", 300 - params_len);
				params_len++;
				strncat(params, in_param,
					300 - params_len);

				mode[*mode_pos] = 'k';
				(*mode_pos)++;
				(*curr_param)++;
			}
		}
		break;

	default:
		break;
	}
}

/*
 * Deals with a mode change for a mode without a parameter
 */
static void chan_mode_change_without_param(ChannelInfo const chan,
					   const int givetake,
					   const char mode_char,
					   char *mode, int *mode_pos,
					   char *params, int *param_len)
{

	switch (mode_char) {

	case 'm':
		if (givetake > 0 && !(chan->mode & CMODE_MODERATED)) {
			mode[*mode_pos] = 'm';
			(*mode_pos)++;
			chan->mode |= CMODE_MODERATED;
		}
		if (givetake < 0 && (chan->mode & CMODE_MODERATED)) {
			mode[*mode_pos] = 'm';
			(*mode_pos)++;
			chan->mode &= ~CMODE_MODERATED;
		}
		break;

	case 'n':
		if (givetake > 0 && !(chan->mode & CMODE_MEMONLYSEND)) {
			mode[*mode_pos] = 'n';
			(*mode_pos)++;
			chan->mode |= CMODE_MEMONLYSEND;
		}
		if (givetake < 0 && (chan->mode & CMODE_MEMONLYSEND)) {
			mode[*mode_pos] = 'n';
			(*mode_pos)++;
			chan->mode &= ~CMODE_MEMONLYSEND;
		}
		break;

	case 't':
		if (givetake > 0 && !(chan->mode & CMODE_TOPIC)) {
			mode[*mode_pos] = 't';
			(*mode_pos)++;
			chan->mode |= CMODE_TOPIC;
		}
		if (givetake < 0 && (chan->mode & CMODE_TOPIC)) {
			mode[*mode_pos] = 't';
			(*mode_pos)++;
			chan->mode &= ~CMODE_TOPIC;
		}
		break;

	case 's':
		if (givetake > 0 && !(chan->mode & CMODE_SECRET)) {
			mode[*mode_pos] = 's';
			(*mode_pos)++;
			chan->mode |= CMODE_SECRET;
		}
		if (givetake < 0 && (chan->mode & CMODE_SECRET)) {
			mode[*mode_pos] = 's';
			(*mode_pos)++;
			chan->mode &= ~CMODE_SECRET;
		}
		break;

	case 'r':
		if (givetake > 0 && !(chan->mode & CMODE_SERVREG)) {
			mode[*mode_pos] = 'r';
			(*mode_pos)++;
			chan->mode |= CMODE_SERVREG;
		}
		if (givetake < 0 && (chan->mode & CMODE_SERVREG)) {
			mode[*mode_pos] = 'r';
			(*mode_pos)++;
			chan->mode &= ~CMODE_SERVREG;
		}
		break;
	case 'R':
		if (givetake > 0 && !(chan->mode & CMODE_REGONLY)) {
			mode[*mode_pos] = 'R';
			(*mode_pos)++;
			chan->mode |= CMODE_REGONLY;
		}
		if (givetake < 0 && (chan->mode & CMODE_REGONLY)) {
			mode[*mode_pos] = 'R';
			(*mode_pos)++;
			chan->mode &= ~CMODE_REGONLY;
		}
		break;
	default:
		break;
	}

}

/*
 * Change a channel mode
 */
static int chan_mode_change(ServerInfo server,
			    ClientInfo sender, ChannelInfo chan, Cmd cmd)
{
	/* Mode string we're gonna reply with */
	char mode[10];
	/* Params in string form for replies */
	char params[300];
	int parm_len = 0;
	/* Output buffer */
	char out_buffer[BUF_SIZE];
	/* Lots of integers */
	int i, param_len, givetake = 0;
	int mode_offset, curr_param;
	ChannelMember chan_user = NULL;
	ClientInfo client = NULL;
	int new_limit = -1;
	int reported_error = 0;
	char *ban_retval = NULL;

	/* Get length of string */
	param_len = strlen(cmd->params[1]);

	/* Set offset to zero in new mode string */
	mode_offset = 0;
	/* Current param */
	curr_param = 2;
	params[0] = '\0';

	/* Go through param and split into parts for our mode string */
	for (i = 0; i < param_len; i++) {
		switch ((cmd->params[1])[i]) {

		case '+':
			if (givetake != 1) {
				givetake = 1;
				mode[mode_offset] = '+';
				mode_offset++;
			}
			break;

		case '-':
			if (givetake != -1) {
				givetake = -1;
				mode[mode_offset] = '-';
				mode_offset++;
			}
			break;


		case 'o':
		case 'h':
		case 'b':
		case 'v':
		case 'l':
		case 'k':
			/* Modes that use a parameter */
			chan_mode_change_with_param(sender, chan, givetake,
						    cmd->params[1][i],
						    cmd->
						    params[curr_param],
						    &curr_param, mode,
						    &mode_offset, params);
			break;
		case 'n':
		case 'm':
		case 't':
		case 's':
		case 'p':
		case 'r':
		case 'R':
			/* Modes without a parameter */
			chan_mode_change_without_param(chan, givetake,
						       cmd->params[1][i],
						       mode,
						       &mode_offset,
						       params, &parm_len);
			break;

		default:
			/* Return an error */
			if (!reported_error) {
				char err_buffer[BUF_SIZE];
				reported_error = 1;
				ircd_sprintf(err_buffer, BUF_SIZE,
					     " %c :Unknown MODE flag",
					     cmd->params[1][i]);
				send_numeric(sender, get_my_server(),
					     501, err_buffer);
			}
			break;

			/* End of case for mode char */
		}
		/* End of for loop to loop though chars in mode string */
	}

	/* Terminate String */
	mode[mode_offset] = '\0';

	if (mode[1] != '\0' && mode[0] != '\0') {
		/* Send to clients */
		ircd_sprintf(out_buffer, BUF_SIZE, "MODE %s %s%s",
			     chan->name, mode, params);
		if (sender->fd != -1)
			send_client_client(sender, sender, out_buffer);
		send_to_local_channel(sender, chan, out_buffer);
		servers_relay(server, sender->nick, out_buffer);
	}

	return 1;

}

/*
* Report the current channel mode
*/
static int chan_mode_report(ClientInfo client, ChannelInfo chan)
{
	char mode[10];
	char params[100];
	char buffer[BUF_SIZE];
	ChannelMember chan_user = NULL;
	int mode_pos = 1;

	/* Are they on that channel */
	chan_user = channel_is_client_on(client, chan);
	mode[0] = '+';
	params[0] = '\0';
	if (chan->mode & CMODE_TOPIC) {
		mode[mode_pos++] = 't';
	}
	if (chan->mode & CMODE_MEMONLYSEND) {
		mode[mode_pos++] = 'n';
	}
	if (chan->mode & CMODE_SECRET) {
		mode[mode_pos++] = 's';
	}
	if (chan->mode & CMODE_MODERATED) {
		mode[mode_pos++] = 'm';
	}
	if (chan->mode & CMODE_SERVREG) {
		mode[mode_pos++] = 'r';
	}
	if (chan->user_limit != -1) {
		mode[mode_pos++] = 'l';
		if (chan_user != NULL) {
			ircd_sprintf(params, BUF_SIZE, "%d",
				     chan->user_limit);
		}
	}
	if (chan->mode & CMODE_KEY) {
		/* Not done */
		mode[mode_pos++] = 'k';
		if (chan_user)
			ircd_sprintf(params, BUF_SIZE, "%s %s", params,
				     chan->key);
	}

	mode[mode_pos] = '\0';

	/* Send mode line */
	ircd_sprintf(buffer, BUF_SIZE, "%s %s %s", chan->name, mode,
		     params);
	send_numeric(client, get_my_server(), 324, buffer);

	return 1;
}

/*
 * Check to see if we can return the channel ban list
 */
static int check_mode_ban(char *modeline)
{
	int i, mode_len;

	mode_len = strlen(modeline);

	for (i = 0; i < mode_len; i++) {
		if (modeline[i] == 'b')
			return 1;

	}
	return 0;
}

/*
 * List the channel ban list
 */
static int list_bans(ClientInfo client, ChannelInfo chan)
{
	int i;
	char buffer[BUF_SIZE];


	for (i = 0; i < chan->ban_size && chan->ban_list[i] != NULL; i++) {
		ircd_sprintf(buffer, BUF_SIZE, "%s %s", chan->name,
			     chan->ban_list[i]);
		send_numeric(client, get_my_server(), 367, buffer);
	}

	ircd_sprintf(buffer, BUF_SIZE, "%s :End of channel ban list",
		     chan->name);
	send_numeric(client, get_my_server(), 368, buffer);

	return 1;
}

/*
 * Deal with a client mode change
 */
static int client_mode(ServerInfo s_ptr, ClientInfo client, char *modeline)
{
	int i;
	char new_mode[10];
	char buffer[100];
	int givetake = 0;
	int mode_pos = 0;

	if (modeline == NULL)
		return 0;


	for (i = 0; modeline[i] != '\0'; i++) {
		switch (modeline[i]) {
		case '+':
			if (givetake != 1) {
				givetake = 1;
				new_mode[mode_pos] = '+';
				mode_pos++;
			}
			break;
		case '-':
			if (givetake != -1) {
				givetake = -1;
				new_mode[mode_pos] = '-';
				mode_pos++;
			}
			break;
		case 'w':
			if (givetake == 1
			    && !(client->mode & MODE_WALLOPS)) {
				/* Give them wallops */
				client->mode |= MODE_WALLOPS;
				new_mode[mode_pos] = 'w';
				mode_pos++;
			}
			if (givetake == -1
			    && (client->mode & MODE_WALLOPS)) {
				client->mode ^= MODE_WALLOPS;
				new_mode[mode_pos] = 'w';
				mode_pos++;
			}
			break;
		case 's':
			if (givetake == 1
			    && !(client->mode & MODE_SERVNOT)) {
				client->mode |= MODE_SERVNOT;
				new_mode[mode_pos] = 's';
				mode_pos++;
			}
			if (givetake == -1
			    && (client->mode & MODE_SERVNOT)) {
				client->mode ^= MODE_SERVNOT;
				new_mode[mode_pos] = 's';
				mode_pos++;
			}
			break;
		case 'i':
			if (givetake == 1 && !(client->mode & MODE_INVIS)) {
				client->mode |= MODE_INVIS;
				new_mode[mode_pos] = 'i';
				mode_pos++;
			}
			if (givetake == -1 && (client->mode & MODE_INVIS)) {
				client->mode ^= MODE_INVIS;
				new_mode[mode_pos] = 'i';
				mode_pos++;
			}
			break;
		case 'a':
			if (givetake == 1 && !(client->mode & MODE_AWAY)) {
				client->mode |= MODE_AWAY;
				new_mode[mode_pos] = 'a';
				mode_pos++;
			}
			if (givetake == -1 && (client->mode & MODE_AWAY)) {
				client->mode ^= MODE_AWAY;
				new_mode[mode_pos] = 'a';
				mode_pos++;
			}
			break;


		case 'o':
			if (givetake == 1 && s_ptr != NULL) {
				/* Do nothing */
				if (!(client->mode & MODE_IRCOP)) {
					ops++;
					client->mode |= MODE_IRCOP;
					new_mode[mode_pos] = 'o';
					mode_pos++;
				}
			}
			if (givetake == -1) {
				if (client->mode & MODE_IRCOP) {
					ops--;
					client->mode ^= MODE_IRCOP;
					new_mode[mode_pos] = 'o';
					mode_pos++;
				}
			}
			break;

		default:
			break;

		}
	}
	new_mode[mode_pos] = '\0';
	if (new_mode[1] != '\0' && new_mode[0] != '\0') {
		ircd_sprintf(buffer, BUF_SIZE, "MODE %s %s", client->nick,
			     new_mode);
		if (client->fd != -1) {
			send_origin_client(client, client->nick, buffer);
		}
		servers_relay(s_ptr, client->nick, buffer);
	}

	return 1;
}

/*
 * Return a client's mode
 */
static int client_mode_return(ClientInfo client)
{
	int mode_pos = 1;
	char curr_mode[12];

	curr_mode[0] = '+';

	if (client->mode & MODE_IRCOP) {
		curr_mode[mode_pos] = 'o';
		mode_pos++;
	}
	if (client->mode & MODE_SERVNOT) {
		curr_mode[mode_pos] = 's';
		mode_pos++;
	}
	if (client->mode & MODE_WALLOPS) {
		curr_mode[mode_pos] = 'w';
		mode_pos++;
	}
	if (client->mode & MODE_INVIS) {
		curr_mode[mode_pos] = 'i';
		mode_pos++;
	}
	if (client->mode & MODE_AWAY) {
		curr_mode[mode_pos] = 'a';
		mode_pos++;
	}
	if (client->mode & MODE_REGIST) {
		curr_mode[mode_pos] = 'r';
		mode_pos++;
	}
	curr_mode[mode_pos] = '\0';

	if (curr_mode[1] == '\0')
		return 1;

	send_numeric(client, get_my_server(), 221, curr_mode);

	return 1;
}
