/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: cmd_nick.c,v 1.34.4.7 2002/04/02 11:33:18 lloydy Exp $
 *
 */
#include "modules.h"

CMIRCD_MODULE(1.02, CMD_CLIENT | CMD_SERVER);

/* Utility routines used in here */
static int nick_taken(ClientInfo, char *);
static int inform_users(ServerInfo, ClientInfo, char *);
static int update_tree(ClientInfo, char *);
static int cmd_nick_remote(ServerInfo, Cmd);

/* 
* Inform client if nick is in use 
*/
static int nick_taken(ClientInfo user_info, char *requested_nick)
{
	char buffer[BUF_SIZE];

	ircd_sprintf(buffer, BUF_SIZE, "%s :Nickname already in use",
		     requested_nick);
	return send_numeric(user_info, get_my_server(), 433, buffer);
}

/* 
* inform room of nick change 
* we need to copy requested_nick into a buffer, 
* and make sure it isn't overlength
*/
static int inform_users(ServerInfo s_ptr, ClientInfo user_info,
			char *requested_nick)
{
	char buffer[BUF_SIZE] = "NICK ";
	char actual_nick[NICK_LENGTH + 1];
	ChannelMember curr_user = NULL;
	int i;
	int rand_no;

	rand_no = rand() % 1000 + 1;
	strncpy(actual_nick, requested_nick, NICK_LENGTH);
	actual_nick[NICK_LENGTH] = '\0';
	ircd_sprintf(buffer, BUF_SIZE, "NICK %s", actual_nick);

	/* Inform client itself */
	if (user_info->fd != -1) {
		/* Don't retell the client, if they are remote */
		send_client_client(user_info, user_info, buffer);
	}

	/* Inform rooms */
	if (user_info->chan_list[0] != NULL) {
		for (i = 0;
		     user_info->chan_list[i] != NULL
		     && i < user_info->chanlist_no; i++) {

			curr_user = user_info->chan_list[i]->users_start;
			while (curr_user) {
				if (curr_user->user_info->rand_no !=
				    rand_no
				    && curr_user->user_info != user_info
				    && curr_user->user_info->fd != -1) {
					send_client_client(curr_user->
							   user_info,
							   user_info,
							   buffer);
					curr_user->user_info->rand_no =
					    rand_no;
				}

				curr_user = curr_user->next;
			}
		}
	}

	/* Inform servers */
	ircd_sprintf(buffer, BUF_SIZE, "NICK %s :%ld", actual_nick,
		     user_info->ts);
	servers_relay(s_ptr, user_info->nick, buffer);

	return 1;
}

/*
* update the client tree
*/
static int update_tree(ClientInfo user_info, char *new_nick)
{

	/* Remove client from tree */
	client_unlink(user_info->nick);

	/* Update various arrays and structs */
	strncpy(user_info->nick, new_nick, NICK_LENGTH);
	user_info->nick[NICK_LENGTH] = '\0';

	/* Add back into tree */
	client_add_by_struct(user_info);
	return 1;
}

/*
 * cmd_nick, checks params and calls correct procedure 
*/
int cmd_nick(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
	char *new_nick = NULL;
	char buffer[BUF_SIZE];

	/* If s_ptr and cmd->origin are non-null its a remote
	   * nick change, and we assume the checks are already done 
	   * Otherwise we check the format of the nick and return an 
	   * error */

	if (s_ptr) {
		/* Its a remote nick command */
		cmd_nick_remote(s_ptr, cmd);

		return 1;
	} else {
		/* From a local client */
		if (cmd->text || cmd->params[0]) {
			/* We have a nick */
			if (cmd->text) {
				new_nick = cmd->text;
			} else {
				new_nick = cmd->params[0];
			}

			/* Check the format of the nick */
			if (!check_nick(new_nick)) {
				ircd_sprintf(buffer, BUF_SIZE,
					     "%s :Erroneous Nick Name",
					     new_nick);
				send_numeric(c_ptr, get_my_server(), 432,
					     buffer);
				return 0;
			}

			/* Check to see if its the same as their current nick */
			if (strncmp(new_nick, c_ptr->nick, NICK_LENGTH) ==
			    0) {
				return 1;
			}

			if ((!client_locate(new_nick)
			     && !temp_locate(new_nick))
			    || lc_strncmp(c_ptr->nick, new_nick,
					  NICK_LENGTH) == 0) {
				/* Change nick and inform everyone */
				c_ptr->ts = time(NULL);
				inform_users(s_ptr, c_ptr, new_nick);
				update_tree(c_ptr, new_nick);
				return 1;
			} else {
				/* Client already has that nick */
				nick_taken(c_ptr, new_nick);
				return 0;

			}
		} else {
			/* Not enough params */
			/* Return numeric */
		}
	}

	return 1;
}

/* 
 * Called by cmd_nick to effect a remote nick change on our
 * tree
 */
static int cmd_nick_remote(ServerInfo s_ptr, Cmd cmd)
{
	ClientInfo c_ptr = NULL;
	ClientInfo n_ptr = NULL;
	ClientInfo e_ptr = NULL;
	ServerInfo a_ptr = NULL, m_ptr = NULL;
	char *new_nick = NULL;
	char buffer[BUF_SIZE];
	int i;

	/* Try and locate client using the origin */
	c_ptr = client_locate(cmd->origin);
	if (!(c_ptr) && !(cmd->params[5])) {
		log_msg(LOG_BASIC,
			"Got a message regarding an existing client, with > 4 params");
		log_msg(LOG_BASIC,
			"This is bad generally, since extended nick messages should only");
		log_msg(LOG_BASIC, "be used when a client connects");
		log_msg(LOG_BASIC,
			"Report this to chris@monkeyircd.org, and try to include some indication of");
		log_msg(LOG_BASIC, "what happened.");
		return 0;
	}

	if (!c_ptr) {

		/* Allocate memory, don't add to tree or announce yet */
		n_ptr = ircd_malloc(sizeof(struct _clientinfo));

		/* Fill out structure */
		n_ptr->fd = -1;
		n_ptr->server = server_locate(cmd->params[6]);
		n_ptr->server->client_count++;
		strncpy(n_ptr->nick, cmd->params[0], NICK_LENGTH);
		strncpy(n_ptr->host, cmd->params[5], HOST_LENGTH);
		strncpy(n_ptr->user, cmd->params[4], USER_LENGTH);
		n_ptr->ts = strtoul(cmd->params[2], NULL, 10);
		n_ptr->idle_timer = time(NULL);
		if (cmd->text) {
			strncpy(n_ptr->real_name, cmd->text,
				REAL_NAME_LENGTH);
		}

		n_ptr->chan_list = calloc(10, sizeof(ChannelInfo));
		n_ptr->chanlist_no = 10;

		/* Hrm, I need a mode routine */
		for (i = 0; cmd->params[3][i] != '\0'; i++) {
			switch (cmd->params[3][i]) {
			case 'w':
				n_ptr->flags |= MODE_WALLOPS;
				break;

			case 'i':
				n_ptr->mode |= MODE_INVIS;
				break;

			case 'r':
				n_ptr->mode |= MODE_REGIST;
				break;

			case 'o':
				n_ptr->mode |= MODE_IRCOP;
				ops++;
				break;

			case 's':
				n_ptr->mode |= MODE_SERVNOT;
				break;

			default:
				break;
			}
		}

		/* Try and check for dups */
		e_ptr = client_locate(n_ptr->nick);
		if (e_ptr) {
			/* We need to get the local server e_ptr is linked via */
			m_ptr = get_my_server();

			a_ptr = e_ptr->server;
			while (a_ptr != NULL && a_ptr->fd == -1) {
				a_ptr = a_ptr->via_ptr;
			}

			/* Pick one to kill */
			if (e_ptr->ts > n_ptr->ts) {
				/* Kill e_ptr */
				client_kill(s_ptr, m_ptr, NULL, e_ptr,
					    "Nick collision, you exist, older nick incoming");
				e_ptr = NULL;
				ircd_sprintf(buffer, BUF_SIZE,
					     "NICK %s %d %ld %s %s %s %s * :%s",
					     n_ptr->nick,
					     n_ptr->server->hops + 1,
					     n_ptr->ts,
					     get_mode_str(n_ptr->mode),
					     n_ptr->user, n_ptr->host,
					     n_ptr->server->name,
					     n_ptr->real_name);
				client_add_by_struct(n_ptr);

			} else {
				client_free(n_ptr);
			}


		} else {
			ircd_sprintf(buffer, BUF_SIZE,
				     "NICK %s %d %ld %s %s %s %s * :%s",
				     n_ptr->nick, n_ptr->server->hops + 1,
				     n_ptr->ts, get_mode_str(n_ptr->mode),
				     n_ptr->user, n_ptr->host,
				     n_ptr->server->name,
				     n_ptr->real_name);
			client_add_by_struct(n_ptr);
		}
		servers_relay(e_ptr ? a_ptr : s_ptr, NULL, buffer);

	} else {
		/* Its a nick change */
		/* Assume client's local server has already checked for clashes */
		/* We should check our list */
		new_nick = cmd->params[0];
		if (cmd->text) {
			c_ptr->ts = strtoul(cmd->text, NULL, 10);
		}
		inform_users(s_ptr, c_ptr, new_nick);
		/* use current contents of client struct before updating them */
		update_tree(c_ptr, new_nick);
	}
	return 1;

}
